//arduino pro or pro mini
#include <stdlib.h>
#include <Math.h>


unsigned char inline isWhiteSpace(char c);
unsigned char inline isAlpha(char c);
unsigned char inline isDigit(char c);
unsigned char inline isAlNum(char c);
char inline toLower(char c);
void inline getChar(void);
void skipWhite(void);
void getRealNumber(void);
void getCmdID(void);
void getName(void);
char readProgByte(const __FlashStringHelper* msg,byte index);

void inline doSteps(void);
void setDir(byte which, char dir);

char gCmdBuffer[256];
unsigned char gCmdBufferPoint=0;
char* gLook=NULL;

//make locals globals for nonrecursive fxns to get good idea of how much dynamic mem is used.
char _gNumbuff[16];
byte _gNumOrNameInc=0;
char _getNumber_sign=1;
byte _getNumber_isfloat=0;
     
unsigned int gCmdID=0;
double gNumber=0.0;
char gName[16];
byte gErrorState=0;

const unsigned char XYSTEPPERENABLE=A0;
const unsigned char ZSTEPPERENABLE=A3;
const unsigned char XAXISDIR=A2;
const unsigned char XAXIS=A1;
const unsigned char ZAXISDIR=10;
const unsigned char ZAXIS=9;
const unsigned char YAXISDIR=11;
const unsigned char YAXIS=12;
const unsigned char PADLEFT=4;
const unsigned char PADRIGHT=5;
const unsigned char PADZUP=2;
const unsigned char PADZDOWN=3;
const unsigned char PADUP=6;
const unsigned char PADDOWN=7;
const unsigned char FANPIN=13;
const unsigned char LEVELPIN=8;

byte gKeyPadEnabled=0;

const byte AXISSTEP[]={XAXIS,YAXIS,ZAXIS};
const byte AXISDIR[]={XAXISDIR,YAXISDIR,ZAXISDIR};

const unsigned int STEPS_PER_XY[2]={80,2032};//0 - millimeters, 1 - inches
const unsigned int STEPS_PER_Z[2]={400,10160};//{400,10160};//0 - millimeters, 1 - inches

long gCurPos[3]={0,0,0};//current x, y, and z positions in step increments.

double gXWord,gYWord,gZWord,gDwellWord,gEWord,gSWord,gTWord;

const double MAX_FEED_RATE=15000;
double gFeedWord=MAX_FEED_RATE;

const byte METRIC=0;
const byte IMPERIAL=1;
byte gUnitType=METRIC;

const byte ABSOLUTE=0;
const byte RELATIVE=1;
byte gCoordType=ABSOLUTE;

const byte UNITSPERMIN=0;
const byte INVERSETIME=1;
byte gFeedType=UNITSPERMIN;

byte inline isWord(char c){
     switch (toLower(c)){
          case 'x':
          case 'y':
          case 'z':
          case 'f':
          case 'p':
          case 'e':
          case 's':
          case 't':
               return 1;
     } 
     return 0;
}

byte loadWord(char wordType){
    if (!isWord(wordType)){
         return 0; 
    }
    getRealNumber();
    switch (toLower(wordType)){
         case 'x': 
              gXWord=gNumber;
              break;
         case 'y':
              gYWord=gNumber;
              break;
         case 'z':
              gZWord=gNumber;
              break;
         case 'f':
              gFeedWord=gNumber;
              break;
         case 'p':
              gDwellWord=gNumber;
              break;
         case 'e':
              gEWord=gNumber;
              break;
         case 's':
              gSWord=gNumber;
              break;
         case 't':
              gTWord=gNumber;
              break;
    }
    return 1;
}
void loadWords(void){
     char wordType;
     while (isWord(*gLook)){
          wordType=*gLook;
          getChar();
          skipWhite();
          if (!loadWord(wordType)){
               break;
          }
     }
}

byte setWord(char wordType){
    if (!isWord(wordType)){
         return 0; 
    }
    getRealNumber();
    switch (toLower(wordType)){
         case 'x': 
              gXWord=gNumber;
              gCurPos[0]=gNumber*(long)STEPS_PER_XY[gUnitType];
              break;
         case 'y':
              gYWord=gNumber;
              gCurPos[1]=gNumber*(long)STEPS_PER_XY[gUnitType];
              break;
         case 'z':
              gZWord=gNumber;
              gCurPos[2]=gNumber*(long)STEPS_PER_Z[gUnitType];
              break;
         default:
              error(F("unknown G92 word"));
              return 0;
    }
    return 1;
}

void setWords(void){
     char wordType;
     while (isWord(*gLook)){
          wordType=*gLook;
          getChar();
          skipWhite();
          if (!setWord(wordType)){
               break;
          }
     }
}

void interpret(void){
  static char cmdType;
  gLook=&gCmdBuffer[0];
  if (*gLook==';') *gLook=0;//the whole line is a comment, null it out and foget bout it.
  skipWhite();
     while (*gLook!=0 && *gLook!=';'){
          cmdType=toLower(*gLook);
          getChar();
          skipWhite();
          switch (cmdType){
               case 'g':
                    if (gErrorState==0){
                        getCmdID();
                        if (gCmdID!=92){//If its 92, dont load all the words.
                          loadWords();
                        }
                        switch (gCmdID){
                             case 0:
                                  if (gFeedType==INVERSETIME){
                                    gotoXYZ(gXWord,gYWord,gZWord,0,1);
                                  }else{
                                    gotoXYZ(gXWord,gYWord,gZWord,MAX_FEED_RATE,1);
                                  }
                                  break;  
                             case 1:
                                  gotoXYZ(gXWord,gYWord,gZWord,gFeedWord,1);
                                  break;  
                             case 4:
                                  delay(gDwellWord*1000);
                                  break;  
                             case 20:
                                  gUnitType=IMPERIAL;
                                  break;  
                             case 21:
                                  gUnitType=METRIC;
                                  break;  
                             case 28:
                                  //TODO: Implement homing
                                  break;
                             case 90:
                                  gCoordType=ABSOLUTE;
                                  break;  
                             case 91:
                                  gCoordType=RELATIVE;
                                  break;
                             case 92:
                                  setWords();
                                  break;  
                             case 93:
                                  gFeedType=INVERSETIME;
                                  break;  
                             case 94:
                                  gFeedType=UNITSPERMIN;
                                  break;
                             default:
                                  error(F("Unsupported G cmd id"));
                                  break;
                        }
                    }else{
                         error(F("Clear error flag before any more commands, use m105")); 
                    }
                    break;
               case 'm':
                    getCmdID();
                    loadWords();
                    switch (gCmdID){
                         case 101://Enable motors
                              enableMotors(); 
                              break;
                         case 102://Disable motors
                              disableMotors();
                              break;
                         case 103://Enable keypad
                              gKeyPadEnabled=1;
                              break;
                         case 104://Disable keypad
                              gKeyPadEnabled=0;
                              break;
                         case 105://Clear error state
                              gErrorState=0;
                              break;
				                 case 200://Wait for user to hit Zup and Zdown simultaneously, then resume
                               Serial.println(F("PAUSED, PRESS ZUP AND ZDOWN TOGETHER TO RESUME"));
                               while (!(!digitalRead(PADZUP) && !digitalRead(PADZDOWN))){
                                checkKeypad(1);
                               }
                               Serial.println(F("ZUP&ZDOWN PRESSED, RESUMING"));
                        	     delay(500);
                               break;
                         default:
                              error(F("Unsupported G cmd id"));
                              break;
	     
                    }
                    //TODO: find out what M codes I care about and include them
                    break;
               default:           
                    if (!loadWord(cmdType)){
                         error(F("Unrecognized command"));
                    } 
                    break;
          }
          if (gErrorState){
               break; 
          }
     }
}

void error(const __FlashStringHelper* msg){
     Serial.print(F("error: "));
     Serial.println(msg);
     gErrorState=1;
}

void setup() {
  // put your setup code here, to run once:
  Serial.begin(115200);//5760 bytes a second? yah, thats plenty fast. can always up it if i want.
  cncInit();
  Serial.println("r");
}

unsigned char getParityByte(char* string){
	int len=strlen(string);
	unsigned char parity=0;
	for (int i=0;i<strlen(string);i++){
		parity+=string[i];
	}
	if (parity==0 || parity==10 || parity==13){
		parity+=100;
	}
	return parity;
}

void loop() {
     static unsigned long lastReadySent=0;
     static unsigned char parityByte=0;
     checkKeypad(0);
     if (Serial.available()>0){
          int readByte=Serial.read();
          if (readByte!=-1){
               if (readByte==10 || readByte==13){
                    if (gCmdBufferPoint>0){
                         gCmdBuffer[gCmdBufferPoint]=0;
                         parityByte=gCmdBuffer[0];
                         if (getParityByte(&gCmdBuffer[1])==parityByte){
                              Serial.print('k');
                              gCmdBuffer[0]=' ';
                              interpret();
                              Serial.println('r'); 
                              lastReadySent=millis();
                              gCmdBufferPoint=0;
                         }else{
                              gCmdBufferPoint=0;
                              Serial.print('b'); 
                         }
                    }else{
                         Serial.println('r'); 
                         lastReadySent=millis();
                    }

               }else{
                    if (gCmdBufferPoint<sizeof(gCmdBuffer)-1){
                         gCmdBuffer[gCmdBufferPoint++]=readByte;
                    }else{
                         error(F("command buffer overflow"));
                         gCmdBufferPoint=0;
                    }
               }
         }
     }
     if ((millis()-lastReadySent>1500 || millis()<lastReadySent)){//ensure a ready is sent at least every 1500ms
          Serial.println('r'); 
          lastReadySent=millis(); 
     }
}

void checkKeypad(int takeItSlow){//Let the operator move errthang around to, makes it easier to zero everything.
    static unsigned int _keypadSpeed=0;
    if (gKeyPadEnabled){
         byte moved=0;
         byte zmoved=0;
         float x=0;
         float y=0;
         float z=0;
         if (!digitalRead(PADUP)){
            y+=0.1;
            moved=1;
          }
          if (!digitalRead(PADDOWN)){
            y-=0.1;
            moved=1;
          }
          if (!digitalRead(PADLEFT)){
            x-=0.1;
            moved=1;
          }
          if (!digitalRead(PADRIGHT)){
            x+=0.1;
            moved=1;
          }
          if (!digitalRead(PADZUP)){
            z+=0.1;
            moved=1;
            zmoved=1;
          }
          if (!digitalRead(PADZDOWN)){
            z-=0.1;
            moved=1;
            zmoved=1;
          }
          if (moved){
              byte oldUnitType=gUnitType;
              byte oldCoordType=gCoordType;
              gUnitType=METRIC;
              gCoordType=RELATIVE;
              if (_keypadSpeed<3000){
                _keypadSpeed+=5;
              }
              
              if (zmoved && takeItSlow){
                 delay(4);
                 z=z/2;
                 if (_keypadSpeed>150){
                    gotoXYZ(x,y,z,100+_keypadSpeed,1); 
                 }
              }else{
                 gotoXYZ(x,y,z,150+_keypadSpeed,1); 
              }
              gUnitType=oldUnitType;
              gCoordType=oldCoordType;
          }else{
               _keypadSpeed=0; 
          }
     }
}



///////CNC Definitions///
void cncInit(void){
  pinMode(FANPIN,OUTPUT);
  digitalWrite(FANPIN,0);
  
  pinMode(XYSTEPPERENABLE,OUTPUT);
  pinMode(ZSTEPPERENABLE,OUTPUT);
  digitalWrite(XYSTEPPERENABLE,1);
  digitalWrite(ZSTEPPERENABLE,1);
  pinMode(XAXIS, OUTPUT);
  pinMode(XAXISDIR, OUTPUT);
  pinMode(YAXIS, OUTPUT);
  pinMode(YAXISDIR, OUTPUT);
  pinMode(ZAXIS, OUTPUT);
  pinMode(ZAXISDIR, OUTPUT);
  
  pinMode(PADLEFT,INPUT);
  pinMode(PADRIGHT,INPUT);
  pinMode(PADZUP,INPUT);
  pinMode(PADZDOWN,INPUT);
  pinMode(PADUP,INPUT);
  pinMode(PADDOWN,INPUT);
  
  digitalWrite(PADLEFT,1);
  digitalWrite(PADRIGHT,1);
  digitalWrite(PADZUP,1);
  digitalWrite(PADZDOWN,1);
  digitalWrite(PADUP,1);
  digitalWrite(PADDOWN,1);
  
  pinMode(LEVELPIN,INPUT);
  digitalWrite(LEVELPIN,1);

}


double inline GetDistance(double x1,double y1,double z1,double x2,double y2,double z2){
    return sqrt(fabs((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1)+(z2-z1)*(z2-z1)));
}

void delayUSeconds(unsigned int time){
     if (time>14000){
          while (time>14000){
               delayMicroseconds(14000);
               time-=14000; 
          }
          delayMicroseconds(time);
     }else{
          delayMicroseconds(time); 
     }
}

void enableMotors(void){
  digitalWrite(FANPIN,1);
  digitalWrite(XYSTEPPERENABLE,0);//enable motors
  digitalWrite(ZSTEPPERENABLE,0);  
}
void disableMotors(void){
  digitalWrite(XYSTEPPERENABLE,1);//disable motors
  digitalWrite(ZSTEPPERENABLE,1);
  digitalWrite(FANPIN,0); 
}

void gotoXYZ(double toX,double toY, double toZ,double feedRate,byte simpleRampOn){
     long tempOrDelta;//used for swapping variables and for holding the dominate delta
     byte i;
     byte j;//a general purpose loop counter, and also the ramping counter.
     long _countdown;
     long _delta[3];
     long _error[2];
     byte _axisList[3];
     char _inc[3];
     unsigned int uiSleepTime; 
     
     if (gCoordType==ABSOLUTE){
          _delta[0]=toX*(long)STEPS_PER_XY[gUnitType]-gCurPos[0];
          _delta[1]=toY*(long)STEPS_PER_XY[gUnitType]-gCurPos[1];
          _delta[2]=toZ*(long)STEPS_PER_Z[gUnitType]-gCurPos[2];
     }else if (gCoordType==RELATIVE){
          _delta[0]=toX*(long)STEPS_PER_XY[gUnitType];
          _delta[1]=toY*(long)STEPS_PER_XY[gUnitType];
          _delta[2]=toZ*(long)STEPS_PER_Z[gUnitType];
     }else{
          error(F("invalid coordinate mode"));
          return; 
     }
     
     double distance=GetDistance(0,0,0,(double)_delta[0]/STEPS_PER_XY[gUnitType],(double)_delta[1]/STEPS_PER_XY[gUnitType],(double)_delta[2]/STEPS_PER_Z[gUnitType]); 
     
     _error[0]=0;
     _error[1]=0;
     for (i=0;i<3;i++){
          _axisList[i]=i;
     }
     for (j=0;j<3;j++){//bubble sort
          for (i=0;i<2;i++){
               if (abs(_delta[i])<abs(_delta[i+1])){
                    tempOrDelta=_delta[i];
                    _delta[i]=_delta[i+1];
                    _delta[i+1]=tempOrDelta;
                    tempOrDelta=_axisList[i];
                    _axisList[i]=_axisList[i+1];
                    _axisList[i+1]=tempOrDelta;
               }      
          }
     }
     
     if (_delta[0]!=0){   
          for (i=0;i<3;i++){//get directions, and make deltas positive
               _inc[i]=1;
               if (_delta[i]<0){
                    _inc[i]=-1;
                    _delta[i]=abs(_delta[i]);
               }
               _delta[i]+=_delta[i];//double it up
               setDir(_axisList[i],_inc[i]);
          }
     
          if (gFeedType==UNITSPERMIN){
               //Calculate for feed units per minute
               uiSleepTime=min(max(160,(60.0f/(feedRate/distance)*1000000.0f)/((double)_delta[0]/2)),65500)-85;
          }else if (gFeedType==INVERSETIME){
               uiSleepTime=min(max(160,(feedRate/((double)_delta[0]/2))*1000000.0f),65500)-85;
          }else{
               error(F("invalid feed type"));
               return; 
          }
          
          if (simpleRampOn){
               if (uiSleepTime<90){
                 j=254; 
               }else{
                 j=15; 
               }
          }
           
          tempOrDelta=_delta[0]/2;//temp is now delta in dominate axis
          _countdown=tempOrDelta;//I think a simple test against 0 and a simple decrement is quicker then comparing the current position versus desired position.
          while (_countdown--){
               gCurPos[_axisList[0]] += _inc[0];//step the dominate axis
               digitalWrite(AXISSTEP[_axisList[0]],1);
               _error[0]+=_delta[1];
               if (_error[0] > tempOrDelta){//step another axis
                    gCurPos[_axisList[1]] += _inc[1];
                    _error[0] -= _delta[0];
                    digitalWrite(AXISSTEP[_axisList[1]],1);
               }
               _error[1]+=_delta[2];
               if (_error[1] > tempOrDelta){//and step the last one
                    gCurPos[_axisList[2]] += _inc[2];
                    _error[1] -= _delta[0];
                    digitalWrite(AXISSTEP[_axisList[2]],1);
               }
               doSteps();
               
               if (simpleRampOn && j>0 && _countdown>254){
                 j-=1;//decrement j to eventually 0, j is the ramping counter
               }else if (simpleRampOn && _countdown<254){
                 j+=1; 
               }
               delayMicroseconds(uiSleepTime+j);
            //   delayUSeconds(uiSleepTime-85+j);
         }
     }
}

void inline doSteps(void){ 
     delayMicroseconds(50);
     digitalWrite(XAXIS,0);
     digitalWrite(YAXIS,0);
     digitalWrite(ZAXIS,0);
}

void setDir(byte which, char dir){
     if (dir>0){
          digitalWrite(AXISDIR[which],1);
     }else if (dir<0){
          digitalWrite(AXISDIR[which],0);
     }
}






















unsigned char inline isWhiteSpace(char c){
     return (c<=32 && c>=1);
}
unsigned char inline isAlpha(char c){
  return ((c>='a' && c<='z') || (c>='A' && c<='Z'));
}
unsigned char inline isDigit(char c){
  return (c>='0' && c<='9'); 
}
unsigned char inline isAlNum(char c){
  return isDigit(c) || isAlpha(c); 
}
char inline toLower(char c){
     if (c>='A' && c<='Z'){
          c-='A';
          c+='a';
     } 
     return c;
}

void inline getChar(void){  
     if (*gLook!=0 && *gLook!=';'){
          gLook++;
     }else if (*gLook==';'){
          *gLook=0; 
     }
}

void skipWhite(void){ 
     while (isWhiteSpace(*gLook)){
          getChar(); 
     }
}

void getRealNumber(void){
     _getNumber_sign=1;
     _getNumber_isfloat=0;
     
     _gNumOrNameInc=0;
     gNumber=0;

     if (*gLook=='-'){
          _getNumber_sign=-1;
          getChar();
     }else if (*gLook=='+'){
          getChar(); 
     }
     while (_gNumOrNameInc<sizeof(_gNumbuff)-1){
          if (isDigit(*gLook) || (_getNumber_isfloat==0 && *gLook=='.')){
               if (*gLook=='.'){
                    _getNumber_isfloat=1;
               }
               _gNumbuff[_gNumOrNameInc++]=*gLook;  
               getChar();
          }else{
               break; 
          }
     }
     if ((_gNumbuff[0]=='.' && _gNumOrNameInc>1) || (_gNumOrNameInc>0 && _gNumbuff[0]!='.')){
          _gNumbuff[_gNumOrNameInc]=0; 
          gNumber=atof(_gNumbuff)*_getNumber_sign;
          skipWhite();
     }else{
          error(F("expected number"));
     }
}
void getCmdID(void){
     gCmdID=0;
     _gNumOrNameInc=0;
     while (_gNumOrNameInc<sizeof(_gNumbuff)-1){
          if (isDigit(*gLook)){
               _gNumbuff[_gNumOrNameInc++]=*gLook; 
               getChar(); 
          }else{
               break; 
          }
     }
     if (_gNumOrNameInc>0){
          _gNumbuff[_gNumOrNameInc]=0;    
          gCmdID=atoi(_gNumbuff);
          skipWhite();
     }else{
          error(F("expected cmd id"));
     }
}

void getName(void){
     _gNumOrNameInc=0;
     gName[0]=0;
     if (isAlpha(*gLook)==0){
          error(F("expected name"));          
     }else{
          while (isAlNum(*gLook)){
              if (_gNumOrNameInc<sizeof(gName)-1){
                   gName[_gNumOrNameInc++]=*gLook;
                   getChar();
              }else{
                   error(F("name buffer overflow"));
                   return;
              }
          }
          gName[_gNumOrNameInc]=0;
         skipWhite(); 
     }
}

char readProgByte(const __FlashStringHelper* msg,byte index){
     return pgm_read_byte(((char*)msg)+index); 
}







/*
//gets strlen up to 254 up a progmem str
byte fstrlen(const __FlashStringHelper* msg){
    byte i=0;
    for (i=0;i<255;i++){
         if (readProgByte(msg,i)==0){
           return i; 
         }
    }
    return -1; 
}

//iCompareStrings, compares case insensitive a progmem str and a ram str, max len of 254.
byte iCompareStrings(const __FlashStringHelper* msg,char const* str){
     byte len=fstrlen(msg);
     if (len!=strlen(str)){
          return 0; 
     }
     while (len-- > 0){
          if (toLower(readProgByte(msg,len))!=toLower(str[len])){
               return 0;
          }
     }
     return 1;
}*/
